'''
Created on 30 июл. 2020 г.

@author: ladmin
'''

import os, stat, re

from abc import ABC, abstractmethod, abstractproperty
from fabric import  colors, api as fab
#from fabric import tasks as taskwrap
from fabricio.tasks import ImageBuildDockerTasks, DockerTasks, Tasks
from fabricio import docker, tasks

from srv.dockerfiles import DockerfileFactory

from fabric.tasks import Task, WrappedCallableTask, get_task_details

from fabric2 import SerialGroup as Group
from fabric2 import Connection

def fname(function_to_decorate):
    def the_wrapper_around_the_original_function():
        print("Я - код, который отработает до вызова функции")
        function_to_decorate() # Сама функция
        print("А я - код, срабатывающий после")
    return the_wrapper_around_the_original_function

class Element(ABC):
    registry=''
    _service=None
    
    dockerlogin=False
    
    @abstractmethod
    def initservice(self):
        pass
    
    def prepareDir(self):
        try:
            os.mkdir(self.layerPath)
            os.chmod(self.layerPath, stat.S_IRWXU, follow_symlinks=True)
        except OSError:
            print ("Создать директорию %s не удалось" % self.layerPath)
        else:
            print ("Успешно создана директория %s " % self.layerPath)
    
    @property
    def opt(self):
        return self.options.copy()
    @opt.setter
    def opt(self, val):
        if 'options' not in self.__dict__:
            self.options={}
        
        try:
            optname, value = val
        except ValueError:
            raise ValueError("Pass an iterable with two items")
        else:
            """ This will run only if no exception was raised """
            self.options[optname]=value
        
    def deleteOpt(self, optname):
        if 'options' not in self.__dict__:
            self.options={}
        try:
            del self.options[optname]
        except Exception as ex:
            print(ex)
            print('Value {optname} is not delete'.format(optname=optname))
    @property
    def network(self):
        return self.__net
    @network.setter
    def network(self, net):
        net.setRole(self._role)
        self.__net=net
        self.opt=('network', [self.__net.netname])
        
    @property
    def volumes(self):
        return  self.__vols
    @volumes.setter
    def volumes(self, vol):
        self.__vols={}
        self.__vols[vol.source]=vol
        self.opt=('mount', vol.generateVolString())
    
    def setService(self, service):
        self._service=service
    
    def add_git_repo(self, gitrepos):       
        self._gitrepos=gitrepos
       
    #@abstractmethod    
    def add_dockerfile(self, dockerfileObj):
        self._dockerfileObj=dockerfileObj
    #@abstractmethod    
    def applay_dockerfile(self):
        self._dockerfileObj.passFileToProject(self.layerPath)
    
    def changeRole(self, role):
        self._role=role
        for obj in vars(self).values():
            if isinstance(obj, WrappedCallableTask):
                obj.roles=role
    
    #@abstractmethod     
    def initRepoForLayer(self, branch='master'):
        self.prepareDir()
        try:
            rep=self._gitrepos.getRepo(workdir=self.layerPath, cur_branch=branch)
        except Exception as ex:
            print(ex)
        return rep
    
    def convertToService(self, tag='v1'):
        service=docker.Service(
            name=self.imageTag,
            image=self.serviceName+':'+tag,
            options=self.opt,
            )
        self.setService(service)
    
    @fab.hosts()
    @fab.roles()
    @fab.task    
    def updateFromImage(self, tag=None, force=False):
        """
        update service use prebiuild image
        """
        fab.execute(self.pull, tag=tag)
        fab.execute(self.update, tag=tag, force=force)
         
    def dockerLogin(self):
        if self.dockerlogin:
            cmd='docker login '+self.registry
            self._role.run(cmd)
              
    @fab.hosts()
    @fab.roles()
    @fab.task
    def commit_task(self, tag='v1'):
        cmd='docker commit {ID} {newServiceTag}'.format(ID=self.service.info['Id'], 
                                                        newServiceTag=self.registry+'/'+self.serviceName+':'+tag)
        fab.run(cmd)
        self.push(tag=tag)
        fab.env.host=None
        fab.env.host_string=None
        #fab.env.passwords=None
    
    @abstractmethod
    def hook(self):
        print("This is hook code")
        
    def getrole(self):
        return self._role
    
class Manager(object):
    def __init__(self):
        pass
        

class NetworkManager(Manager):
    def __init__(self, netname, role=False,  driver='bridge'):
        self.role=role
        self.netname=netname
        self.driver=driver
    def setRole(self, role):
        self.role=role
    def createNetwork(self):
        res=self.role.execute(self._createNetwork)
        return res
    
    def _createNetwork(self, connection):
        try:
            net=connection.networks.get(self.netname)
            print("Network {name} exist".format(name=net))
        except Exception as Ex:
            res=connection.networks.create(self.netname, driver=self.driver)
            res
        return False
        
    def update(self):
        print("updating network")
        self.createNetwork()

class RoleManager(list):
    def __new__(cls, role):
        cls.role=role
        cls.portIndex=23750
        return super().__new__(cls)
    #def __init__(self, role, addr):
        #self.role=role
        #self.addr=addr
    def setAddr(self, addr):
        self.addr=addr    
    
    def execute(self, callFunc):
        
        import paramiko
        from docker import DockerClient
        from sshtunnel import SSHTunnelForwarder
        import time

        for host in self.adress:
            user=host.split('@', 1)[0]
            addr=host.split('@', 1)[1]
            port=self.localConnectionPort
            forward=SSHTunnelForwarder((addr, 22), 
                                ssh_username=user, 
                                ssh_pkey="/var/ssh/rsa_key", 
                                ssh_private_key_password="secret", 
                                remote_bind_address=(addr, 2375), 
                                local_bind_address=('127.0.0.1', port))
            forward.start()
            time.sleep(3)
            dockerConection=DockerClient('tcp://127.0.0.1:{port}'.format(port=port))
            res=callFunc(dockerConection)
            dockerConection.close()
            del dockerConection
            forward.stop()
        return res
    
    @property
    def localConnectionPort(self):
        port=self.portIndex
        self.portIndex+=1
        return port
    
    def getConnections(self):
        dockerConections={}
        for host in self.adress:
            dockerConections[host]=self.openDockerConnection(host=host)
        return dockerConections
    @property
    def adress(self):
        return self.addr
    @adress.setter
    def adress(self, addr):
        self.addr=addr
        
    @property
    def roles(self):
        return self.role
    @roles.setter
    def roles(self, rolename):
        self.role=rolename
        #self.updateGlobalRoles(rolename)
    #@role.deleter
    #def roles(self, rolename):
    #    del self.role_list.remove(rolename)    
    def run(self, arg):
        for host in self.adress:
            result = Connection(host).run(arg, pty=True)
            print("{}: {}".format(host, result.stdout.strip()))
    
    
    
class VolumeManager(Manager):
    def __init__(self, role, typeVol='volume', driver='local', source='defvol', options='readonly=false'):
        self.role=role
        self.driver=driver
        self.source=source
        self.typeVol=typeVol
        self.options=options
        self.template='type={type},source={source},destination={destination},{options}'
    def setDestination(self, destination):
        self.destination=destination
    def setSource(self, source):
        self.source=source
    def setType(self, typeVol):
        self.typeVol=typeVol
    def setAditionalOpt(self, options):
        self.options=options
    
    def update(self):
        res=self.role.execute(self._create)
        return res
    
    def _create(self, connection):
        try:
            vol=connection.volumes.get(self.source)
            print("Volume {name} exist".format(name=vol))
        except Exception as Ex:
            res=connection.volumes.create(self.source, driver=self.driver)
            res
        return False
    
    def generateVolString(self):
        opt=self.template.format(type=self.typeVol, source=self.source, destination=self.destination, options=self.options)
        return [opt]
        


class BlackBox(ABC):
    '''
    classdocs
    '''
    stack={}
    __index=0
    
    projectName=None
    
    def __init__(self):
        '''
        Constructor
        '''
        pass
    #@abstractmethod    
    #def build_layer(self) -> None:
    #    pass
    
    @property
    def layer(self):
        return self.list
    
    @layer.setter
    def layer(self, strategy):
        
        self.stack[strategy._element.imageTag]=strategy
        self._increment()
        
    def _increment(self):
        self.__index+=10
    
    @fab.hosts()
    @fab.roles()
    @fab.task    
    def compile_layers(self):
        for item in self.stack.keys():
            self.stack[item].fullcycle()
    @fab.hosts()
    @fab.roles()
    @fab.task    
    def initAllRepo(self):
        for item in self.stack.keys():
            self.stack[item]._element.runrepo()
    
    @fab.hosts()
    @fab.roles()
    @fab.task    
    def push_layers(self):
        print('==== PUSH ====')



    

class LayerBuilder(ImageBuildDockerTasks, Tasks, Element):
    
    def __init__(self,
                net='network',
                role=['defrole'], 
                dockerlogin=False, registry='',
                projectname='testproj', 
                imageTag='unitname' , 
                release='latest'):
          
        self.dockerlogin=dockerlogin
        self.myname=projectname
        self.registry=registry
        self.imageTag=imageTag
        self.serviceName=projectname+'/'+imageTag
        self.serviceTag=projectname+'/'+imageTag+':'+release
        self.layerPath=os.path.join(os.getcwd(), 'build/'+self.serviceName)
        self._role=role
        self.network=net
        self.destroy = self.DestroyTask(tasks=self)
        
        
    def initservice(self):
        super(LayerBuilder, self).__init__(self._service,
                                      build_path='build/'+self.serviceName,
                                      roles=self._role,
                                      registry=self.registry,
                                      destroy_command=True)
        self.dockerLogin()
    
        
    def commit(self, tag='v1'):
        fab.execute(self.commit_task, tag)

    def hook(self):
        print('rewritd hook')
        
    #@fab.hosts()
    #@fab.roles()
    #@fab.task
    #def deploy(self, tag=None, force=False, backup=False, migrate=True):
    #    """
    #    deploy service (prepare -> push -> backup -> pull -> migrate -> update)
    #    """
    #    self.prepare(tag=tag)
    #    self.push(tag=tag)
    #    self.registry
    #    fab.execute(
    #        self.migrate,
    #        tag=tag,
    #        #force=False,
    #        #backup=False,
    #        #migrate=True,
    #    )
        

class LayerConstructor(DockerTasks, Element):

    def __init__(self,
                net='network',
                role=['defrole'], 
                dockerlogin=False, registry='',
                projectname='testproj', 
                imageTag='unitname' , 
                release='latest'):
        self.dockerlogin=dockerlogin
        self.myname=projectname
        self.registry=registry
        self.imageTag=imageTag
        self.serviceName=projectname+'/'+imageTag
        self.serviceTag=imageTag+':'+release
        self.layerPath=os.path.join(os.getcwd(), self.serviceName)
        self._role=role
        self.network=net
        self.destroy = self.DestroyTask(tasks=self)
    
        
    def initservice(self):
        super(LayerConstructor, self).__init__(self._service,
                                               roles=self._role,
                                               registry=self.registry)
        self.dockerLogin()
             
    def hook(self):
        print('rewritd hook')


class LayerFactorySkeel(ABC):
    dockerFactory=DockerfileFactory()
    
    def __init__(self) -> None:
        self._element = None
        self.__index=0
         
    def reset(self):
        self._element=None
    def __increment(self):
        self.__index+=1
    
    @property
    def builder(self) -> Element:
        return self._element

    @builder.setter
    def builder(self, element: Element) -> None:
        """
        Директор работает с любым экземпляром строителя, который передаётся ему
        клиентским кодом. Таким образом, клиентский код может изменить конечный
        тип вновь собираемого продукта.
        """
        self._element = element

    


   



'''
Created on 30 июл. 2020 г.

@author: ladmin
'''

import os, stat, shutil, re, sys
import json
from fabric import  colors, api as fab
from git import Repo
from git import Git
from git.exc import GitCommandError
from fabric2 import Connection
#from _ast import Try
    

class Dockerfile():
    
    _dockerfile='Dockerfile'
    destFile='Dockerfile'
    _file=None
    _w_dir='/tmp'
    dockerfilePath=os.path.join(_w_dir, _dockerfile)
    
    def __init__(self, layerPath):
        self.layerPath=layerPath
        self.reset()
        #self.open()
        self._index=10
        
    
    def open(self):
        
        try:
            os.remove(self.dockerfilePath)
            self._file = open(self.dockerfilePath, 'w')
            # Do something with the file
        except IOError:
            self._file = open(self.dockerfilePath, 'w')
    
    def close(self):
        self._file.close()
        
    def reset(self):
        self.dockerfile={}
    
    def _increment(self):
        self._index+=10
        
    def buildfile(self):
        dest=os.path.join(self.layerPath, self.destFile)
        file=None
        try:
            os.remove(dest)
            file = open(dest, 'w')
            # Do something with the file
        except IOError:
            file = open(dest, 'w') 
           
        for key in sorted(self.dockerfile.keys()):
            file.write(self.dockerfile[key]+'\n')
        
        file.close()
        print(self.layerPath)
    
    def copyall(self, src, dst, symlinks=False, ignore=None):
        
        if os.path.exists(dst):
            if os.path.isfile(dst) or os.path.islink(dst):
                os.remove(dst)
            else:
                shutil.rmtree(dst, ignore_errors=True, onerror=True)   
        if os.path.isdir(src):
            shutil.copytree(src, dst, symlinks, ignore)
        else:
            shutil.copy2(src, dst)
    
    def add_STRING(self, str_content, index=None):  
        if index is not None:
            self.dockerfile[index]=str_content
        else: 
            self.dockerfile[self._index]=str_content
        self._increment()
        self.ADD.__name__
    def FROM(self, str_content):
        self._file.write("FROM "+str_content+'\n')   
    def RUN(self, str_content):
        self._file.write("RUN "+str_content+'\n')
    def ADD(self, source, destination, index=None):
        if index is not None:
            self.dockerfile[index]=self.ADD.__name__+' '+source+' '+destination
        else:
            self.dockerfile[self._index]=self.ADD.__name__+' '+source+' '+destination
        #print(os.getcwd())
        os.makedirs(os.path.dirname(os.path.join(self.layerPath, source)), stat.S_IRWXU, exist_ok=True)
        self.copyall(source, os.path.join(self.layerPath, source))
        self._increment()
    def EXPOSE(self, str_content):
        self._file.write("EXPOSE "+str_content+'\n')
    def VOLUME(self, str_content):
        self._file.write("VOLUME "+str_content+'\n')
    def CMD(self, str_content):
        self._file.write("CMD "+str_content+'\n')
    def WORKDIR(self, str_content):
        self._file.write("WORKDIR "+str_content+'\n')
    def COPY(self, source, destination, index=None):
        if index is not None:
            self.dockerfile[index]=self.COPY.__name__+' '+source+' '+destination
        else:
            self.dockerfile[self._index]=self.COPY.__name__+' '+source+' '+destination
            
        #print(os.path.join(self.layerPath, source))
        os.makedirs(os.path.dirname(os.path.join(self.layerPath, source)), stat.S_IRWXU, exist_ok=True)
        self.copyall(source, os.path.join(self.layerPath, source))
        self._increment()


class gitworker(object):
    '''
    Класс, реализующий доступ к репозиторию проекта
    '''
    gitrep=0

    def __init__(self, git_rep, keepath=os.getcwd()+'id_deployment_key'):
        '''
        Constructor
        '''
        'ladmin@127.0.0.1:/git/itcluster/.git'
        self.gitrep=git_rep
        self.keepath=keepath
        #self.workdir='/home/ladmin/workspace_cpp/fabrica'   
        
    
    def prepareRepo(self, workdir, cur_branch='master'):     
        try:
            cur_repo = Repo(os.path.join(workdir, 'html'))
            assert not cur_repo.bare
        except Exception as exRepo:
            ssh_cmd='ssh -i '+self.keepath
            #ssh_executable = os.path.join(keepath, self.wrapper)
            print(exRepo)
            #with  Git().custom_environment(GIT_SSH=ssh_executable):
            try:
                with  Git().custom_environment(GIT_SSH_COMMAND=ssh_cmd):
                    cur_repo = Repo.clone_from(self.gitrep, os.path.join(workdir, 'html'), branch=cur_branch)        
            except Exception as exGit:
                print(exGit)
        assert not cur_repo.is_dirty()
        return cur_repo

    def changeBranch(self, repo, origin='origin', branch='master'):
        try:
            repo.git.checkout(branch)
            origin=repo.remote(origin)
            res=origin.pull(branch)
        except Exception as ex:
            print(ex)
        return repo
        
    def getRepo(self, workdir, cur_branch):
        repo=self.prepareRepo(workdir)
        try:
            self.changeBranch(repo=repo, branch=cur_branch)      
        except GitCommandError as ex:
            print(ex)
            raise SystemExit(0)
        return repo

class Networker():
    def __init__(self, netname):
        self._netname=netname
        try:
            inspectStr='docker network inspect '+self._netname
            ret = fab.run(inspectStr)
            self.net_param=json.loads(ret)[0]
            #name = net_param.get("Name")
        except:
            print("Netname not defined. Creating new nerwork ")
            
    def getRelateIpList(self, filter="bridge"):
        containers=self.net_param['Containers']
        iplist={}
        for cont in containers.keys():
            container=containers[cont]
            for param in container.keys():
                if re.match(filter, container[param]) is not None:
                    iplist[container['Name']]=container['IPv4Address'][0:-3]
        
        return iplist


'''
Created on 31 июл. 2020 г.

@author: ladmin
'''
import os, stat, re
from fabric import  colors, api as fab
from abc import ABC, abstractmethod, abstractproperty
from srv.objectlib import Networker

class ExecStrategy(ABC):
    
    def __init__(self, element):
        self._element=element
    def updateRole(self):
        print('Update Role')
        str=' '
        fab.env.host_string=str.join(fab.env.roledefs[str.join(self._element._role)])
      
    @abstractmethod
    def fullcycle(self):
        self.prepare()
    
    @abstractmethod
    def prepare(self):
        self._element.runrepo()  
    @abstractmethod
    def initRepo(self):
        pass
    
    def accompaniment(self):
        self.makeNetwork()
        self.makeVolume()   
    
    def migrate(self, tag=None):
        self._element.initservice()
        self.accompaniment()
        self._element.updateFromImage(force=True, tag=tag)
         
    def makeVolume(self):
        if hasattr(self._element, '_Element__vols'):
            vols=self._element.volumes
            for vol in vols:
                vols[vol].update()
    
    def makeNetwork(self):
        if hasattr(self._element, '_Element__net'):
            self._element.network.update()

class ConstructStrategy(ExecStrategy):
    def initRepo(self, branch='master'):
        repo=self._element.initRepoForLayer(branch)
        return repo
    
    def prepare(self):
        self.accompaniment()
        self._element.hook()
        self._element.initservice()
 
        
    def fullcycle(self):
        self.prepare()
        self._element.deploy()
        fab.env.host=None
        fab.env.host_string=None
        
class CommonStrategy(ExecStrategy):
    def initRepo(self, branch='master'):
        repo=self._element.initRepoForLayer(branch)
        return repo
    def prepare(self):
        self.accompaniment()
        self._element.hook()
        self._element._dockerfileObj.buildfile()
        self._element.initservice()
    
    def fullcycle(self, force=True):
        self.prepare()
        
        #self._element.push()
        #self._element.backup()
        #self._element.pull()
        #self._element.migrate()
        #self._element.update()
        #self._element.upgrade()
        
        self._element.deploy()
        fab.env.host=None
        fab.env.host_string=None
        
        
class CommonStrategyForNginx(ExecStrategy):
    
    def initRepo(self, branch='master'):
        repo=self._element.initRepoForLayer(branch)
        return repo
    
    def prepare(self):
        self.accompaniment()
        self._element.hook()
        self._element._dockerfileObj.buildfile()
        self._element.initservice() 
    
    def fullcycle(self):
        #tag=None
        #force=False
        #backup=False
        #migrate=True
        
        self.prepare()
        #self._element.deploy2(tag='127.0.0.1:5000/itcluster/nginx:latest')
        self._element.deploy()
        #self._element.prepare()
        #self._element.push()
        #fab.execute(
        #    self._element.upgrade,
        #    tag=tag,
        #    force=force,
        #    backup=backup,
        #    migrate=migrate,
        #)
        fab.env.host=None
        fab.env.host_string=None
        
    def putStream(self, streamname, content):
        fullname=os.path.join(os.getcwd(), streamname+'.conf')
        try:
            os.remove(fullname)
            file = open(fullname, 'w')
            # Do something with the file
        except IOError:
            file = open(fullname, 'w')
        file.write(content)   
        file.close()
        self._element._dockerfileObj.ADD(streamname+'.conf', '/etc/nginx/conf.d/'+streamname+'.conf', index=35)
        
        
        
class CommonStrategyForPhp(ExecStrategy):
    
    def prepare(self):
        self.accompaniment()
        self._element.hook()
        self._element._dockerfileObj.buildfile()
        self._element.initservice() 
    
    def fullcycle(self):    
        self.prepare()
        self._element.deploy()
        fab.env.host=None
        fab.env.host_string=None
        
    def initRepo(self, branch='master'):
        repo=self._element.initRepoForLayer(branch)
        return repo

    def getStreamConfig(self):
        self.updateRole()
        f=fab.env
        str=' '

        netId=self._element._service.info['Spec']['TaskTemplate']['Networks']
        net=self._element._service.network
        print(net)
        netName=str.join(net)
        networker=Networker(netName)
        iplist=networker.getRelateIpList(filter=self._element.myname)
        template='upstream '+self._element.myname+' {\n'
        for curName in iplist.keys():
            template=template+'server '+iplist[curName]+':9000 weight=2 max_fails=2 fail_timeout=2s;\n'   
        template=template+'}'

        print(template)
        print('sdafadsfa')
        return template




'''
Created on 8 окт. 2020 г.

@author: ladmin
'''
from setuptools import setup, find_packages

setup(name='fabula-py',
      version='0.0.2',
      description='Framefork for description infrastructure',
      long_description='Framefork for description infrastructure',
      long_description_content_type="text/markdown",
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
      keywords='fabula2',
      url='https://github.com/djmoto24/fabula.git',
      author='skytrain',
      author_email='no067@mail.ru',
      license='MIT',
      packages=find_packages(),
      #install_requires=[
      #    'fabricio',
      #],
      include_package_data=True,
      zip_safe=False)